# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import csv
from pathlib import Path
from typing import Dict, List, TYPE_CHECKING
from dataclasses import dataclass
from collections import OrderedDict

try:
    import pandas as pd
    import plotly.graph_objects as go

    _plotting_pkgs_available = True
except ImportError:
    _plotting_pkgs_available = False

try:
    from IPython import get_ipython
    from IPython.display import Image, display

    _jupyter_available = True
except ImportError:
    get_ipython = None
    Image = None
    display = None
    _jupyter_available = False

from chipscopy.dm import request
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.api import CoreType
from chipscopy.api._detail.debug_core import DebugCore
from chipscopy.utils.printer import PercentProgressBar, printer

if TYPE_CHECKING:
    from chipscopy.client.ddrmc_client import DDRMCClient


@dataclass
class DDR(DebugCore["DDRMCClient"]):
    """
    This class contains the top level API functions to interact with the
    integrated DDR memory controller debug core(s) on Versal devices.

    """

    def __init__(self, ddr_node):
        super(DDR, self).__init__(CoreType.DDR, ddr_node)
        self.node_name = ddr_node.props["Name"]
        self.mc_index = ddr_node.props["mc_index"]
        self.mc_loc = ddr_node.props["mc_loc"]
        self.name = "ddr_" + self.mc_index
        self.ddr_node = self.core_tcf_node
        self.is_enabled = self.is_user_enabled()
        self.is_gen5 = False
        arch_type = ddr_node.props.get("arch")
        if arch_type:
            if arch_type == "gen5":
                self.is_gen5 = True
                self.crypto_en = False
                self.crypto_type = "N/A"
                self.__get_crypto_status()
        self.eye_scan_data = []
        self.filter_by = {"name": self.name, "mc_index": self.mc_index, "mc_loc": self.mc_loc}

    def get_cal_status(self):
        """
        Get the main calibration status of DDRMC

        Args:

        Returns:
            status in string

        """

        status = self.ddr_node.get_cal_status()

        return status

    def refresh_cal_status(self, done: DoneHWCommand = None):
        """
        Refresh the main calibration status of DDRMC

        Args:
            done: Optional command callback that will be invoked when the response is received

        Returns:

        """

        self.ddr_node.refresh_cal_status(done)

    def get_cal_stages(self):
        """
        Get the decoded calibration stages results happened from a
        calibration run

        Args:

        Returns:
            A dictionary where
                Key = calibration stage name
                Val = stage calibration result

        """
        results = {}
        results = self.ddr_node.get_property_group(["stages"])

        return results

    def refresh_cal_margin(self, done: DoneHWCommand = None):
        """
        Refresh the calibration windows margin analysis of DDRMC

        Args:
            done: Optional command callback that will be invoked when the response is received

        Returns:

        """

        self.ddr_node.refresh_cal_margin(done)

    def refresh_health_status(self, done: DoneHWCommand = None):
        """
        Refresh the overall health status analysis of DDRMC

        Args:
            done: Optional command callback that will be invoked when the response is received

        Returns:

        """

        self.ddr_node.refresh_health_status(done)

    def get_cal_margin_mode(self):
        """
        Get the availability of windows margin mode generated from a
        calibration run

        Args:

        Returns:
            A dictionary where
                Key = name of margin mode
                Val = True or False value indicating the availability

        """
        results = {}

        results = self.ddr_node.get_cal_margin_mode()

        return results

    def is_user_enabled(self):
        """
        Find out whether the DDRMC core is user enabled or not after
        Versal device configuration

        Args:

        Returns:
            Status value in Bool

        """
        # if not being powered by NoC clock, it is automatically disabled
        clock_off = self.ddr_node.is_clock_gated()

        if clock_off:
            return False
        else:
            result = {}
            key = "pcsr_control_ub_init"
            result = self.ddr_node.get_property(key)

            if not result[key]:
                return False
            else:
                return True

    def get_property(self, property_names=None, done: DoneHWCommand = None):
        """
        Get the property value mapped to DDRMC core

        Args:
            property_names: single string or list of string of property names
                (if None specified, all properties available will be queried and returned)

            done: Optional command callback that will be invoked when the response is received

        Returns:
            A dictionary, where
                Key = property name
                Val = property value

        """
        results = {}

        results = self.ddr_node.get_property(property_names, done)

        return results

    def __get_crypto_status(self):
        # Only care if it is Gen5 and User enabled
        if not self.is_enabled:
            return

        props = ["crypto_en", "crypto_status"]
        results = self.ddr_node.get_property(props)

        if results["crypto_en"]:
            self.crypto_en = True

        self.crypto_type = results["crypto_status"]

    def __get_configuration(self) -> OrderedDict:
        configs = OrderedDict()
        is_lrdimm = False

        configs["DDRMC Core Name"] = self.name

        results = self.ddr_node.get_property("mem_type")
        val = results["mem_type"]
        if val == 1:
            configs["Memory Interface"] = "DDR4"
        elif val == 2:
            configs["Memory Interface"] = "LPDDR4"
        elif val == 4:
            configs["Memory Interface"] = "DDR5"
        elif val == 5:
            configs["Memory Interface"] = "LPDDR5"
        else:
            configs["Memory Interface"] = "Unknown"

        results = self.ddr_node.get_property("dimm_type")
        val = results["dimm_type"]
        if val == 0:
            configs["Device Type"] = "Component"
        elif val == 1:
            configs["Device Type"] = "UDIMM"
        elif val == 2:
            configs["Device Type"] = "RDIMM"
        elif val == 3:
            configs["Device Type"] = "LRDIMM"
            is_lrdimm = True
        else:
            configs["Device Type"] = "Unknown"

        configs["MC Location"] = self.mc_loc

        results = self.ddr_node.get_property("slots")
        slots = results["slots"]
        results = self.ddr_node.get_property("phy_ranks")
        phy_ranks = results["phy_ranks"]
        results = self.ddr_node.get_property("mem_ranks")
        mem_ranks = results["mem_ranks"]

        if is_lrdimm:
            configs["Slots"] = str(phy_ranks)
            configs["Ranks"] = str(mem_ranks)
        else:
            configs["Slots"] = str(slots)
            configs["Ranks"] = str(phy_ranks)

        results = self.ddr_node.get_property("bytes")
        bytes = results["bytes"]
        results = self.ddr_node.get_property("nibbles")
        nibbles = results["nibbles"]
        results = self.ddr_node.get_property("bits_per_byte")
        bits = results["bits_per_byte"]
        data_width = bytes * bits
        configs["Data Width"] = str(data_width)
        configs["Bytes"] = str(bytes)
        configs["Nibbles"] = str(nibbles)
        configs["Bits per Byte"] = str(bits)

        freq_zero = "Unknown"
        freq_one = "Unknown"
        results = self.ddr_node.get_property("dual_freq_en")
        dual_freq = results["dual_freq_en"]
        if self.is_gen5:
            f0_ck_name = "f0_ck_period"
            f1_ck_name = "f1_ck_period"
        else:
            f0_ck_name = "f0_period"
            f1_ck_name = "f1_period"
        results = self.ddr_node.get_property(f0_ck_name)
        period = results[f0_ck_name]
        if (period != "0") and (period != ""):
            freq0 = int(1000000 / float(period) + 0.5)
            freq_zero = str(freq0) + " MHz"
        configs["Memory Frequency 0"] = freq_zero
        if int(dual_freq) > 0:
            results = self.ddr_node.get_property(f1_ck_name)
            period = results[f1_ck_name]
            if (period != "0") and (period != ""):
                freq1 = int(1000000 / float(period) + 0.5)
                freq_one = str(freq1) + " MHz"
            configs["Memory Frequency 1"] = freq_one

        return configs

    def __get_cal_margins(
        self, mode_base: str, left_margs: Dict, right_margs: Dict, centers: Dict, prbs: bool = False
    ):
        group_name = mode_base + "_left"
        results = self.ddr_node.get_property_group([group_name])
        left_margs.clear()
        left_margs.update(results)
        group_name = mode_base + "_right"
        results = self.ddr_node.get_property_group([group_name])
        right_margs.clear()
        right_margs.update(results)
        centers.clear()
        if not prbs:
            group_name = mode_base + "_center"
            results = self.ddr_node.get_property_group([group_name])
            centers.update(results)

    def __output_read_margins(
        self,
        by_8_mode,
        bytes: int,
        nibbles: int,
        left_margs: Dict,
        right_margs: Dict,
        centers: Dict,
    ):
        left_marg_vals = sorted(left_margs.items())
        center_vals = sorted(centers.items())
        right_marg_vals = sorted(right_margs.items())

        if by_8_mode:
            for byte in range(bytes):
                left_marg = left_marg_vals[byte * 2][1]
                center = center_vals[byte * 2][1]
                right_marg = right_marg_vals[byte * 2][1]
                printer(
                    "Byte ",
                    str(byte),
                    " Nibble 0  -",
                    "  Left Margin:  ",
                    str(left_marg[1]),
                    " (",
                    str(left_marg[0]),
                    ")  Center Point:  ",
                    str(center[1]),
                    " (",
                    str(center[0]),
                    ")  Right Margin:  ",
                    str(right_marg[1]),
                    " (",
                    str(right_marg[0]),
                    ")",
                )

                left_marg = left_marg_vals[byte * 2 + 1][1]
                center = center_vals[byte * 2 + 1][1]
                right_marg = right_marg_vals[byte * 2 + 1][1]
                printer(
                    "Byte ",
                    str(byte),
                    " Nibble 1  -",
                    "  Left Margin:  ",
                    str(left_marg[1]),
                    " (",
                    str(left_marg[0]),
                    ")  Center Point:  ",
                    str(center[1]),
                    " (",
                    str(center[0]),
                    ")  Right Margin:  ",
                    str(right_marg[1]),
                    " (",
                    str(right_marg[0]),
                    ")",
                )
        else:
            for nibble in range(nibbles):
                left_marg = left_marg_vals[nibble][1]
                center = center_vals[nibble][1]
                right_marg = right_marg_vals[nibble][1]
                printer(
                    "Nibble ",
                    str(nibble),
                    "  -" "  Left Margin:  ",
                    str(left_marg[1]),
                    " (",
                    str(left_marg[0]),
                    ")  Center Point:  ",
                    str(center[1]),
                    " (",
                    str(center[0]),
                    ")  Right Margin:  ",
                    str(right_marg[1]),
                    " (",
                    str(right_marg[0]),
                    ")",
                )

    def __output_gen5_margins(
        self,
        is_dq: bool,
        bytes: int,
        bits_byte: int,
        left_margs: Dict,
        right_margs: Dict,
        centers: Dict,
    ):
        left_marg_vals = sorted(left_margs.items())
        right_marg_vals = sorted(right_margs.items())
        center_vals = {}
        if not centers:
            for idx in range(len(left_margs)):
                center_vals[idx] = (0, 0)
        else:
            center_vals = sorted(centers.items())

        if is_dq:
            for byte in range(bytes):
                for bit in range(bits_byte):
                    left_marg = left_marg_vals[byte * bits_byte + bit][1]
                    center = center_vals[byte * bits_byte + bit][1]
                    right_marg = right_marg_vals[byte * bits_byte + bit][1]
                    printer(
                        "Byte ",
                        str(byte),
                        " Bit ",
                        str(bit),
                        " - Left Margin: ",
                        str(left_marg[1]),
                        "(",
                        str(left_marg[0]),
                        ") Center Point: ",
                        str(center[1]),
                        "(",
                        str(center[0]),
                        ") Right Margin: ",
                        str(right_marg[1]),
                        "(",
                        str(right_marg[0]),
                        ")",
                    )
        else:
            # DBI for LP5 Read
            # DM for Write
            for byte in range(bytes):
                left_marg = left_marg_vals[byte][1]
                center = center_vals[byte][1]
                right_marg = right_marg_vals[byte][1]
                printer(
                    "Byte ",
                    str(byte),
                    " - Left Margin: ",
                    str(left_marg[1]),
                    "(",
                    str(left_marg[0]),
                    ") Center Point: ",
                    str(center[1]),
                    "(",
                    str(center[0]),
                    ") Right Margin: ",
                    str(right_marg[1]),
                    "(",
                    str(right_marg[0]),
                    ")",
                )

    def __output_write_margins(
        self, bytes: int, left_margs: Dict, right_margs: Dict, centers: Dict
    ):
        left_marg_vals = sorted(left_margs.items())
        center_vals = sorted(centers.items())
        right_marg_vals = sorted(right_margs.items())

        for byte in range(bytes):
            left_marg = left_marg_vals[byte][1]
            center = center_vals[byte][1]
            right_marg = right_marg_vals[byte][1]
            printer(
                "Byte ",
                str(byte),
                "  -" "  Left Margin:  ",
                str(left_marg[1]),
                " (",
                str(left_marg[0]),
                ")  Center Point:  ",
                str(center[1]),
                " (",
                str(center[0]),
                ")  Right Margin:  ",
                str(right_marg[1]),
                " (",
                str(right_marg[0]),
                ")",
            )

    def report(self, to_file=False, file_name=None):
        """
        Run a report on the current statuses and analytical data of the DDRMC

        Args:
            to_file: specify True to have the report saved to a file

            file_name: the file name or full path for the report file to be saved

        Returns:

        """
        out_file = None
        orig_out = None

        if to_file:
            if not file_name:
                printer("\nNOTE: Need to specify a file name for the report output.")
            else:
                out_file = open(file_name, "w")
                printer("\nNOTE: The report is being generated and saved as:", file_name)
                orig_out = sys.stdout
                sys.stdout = out_file

        # DDRMC Status & Message
        printer("-------------------\n")
        printer(" DDRMC Status \n")
        printer("-------------------\n")

        # User enabled status check
        if not self.is_enabled:
            printer(self.name, "is not enabled.", "\n")
            if out_file:
                sys.stdout = orig_out
                out_file.close()
            return

        self.refresh_cal_status()
        self.refresh_health_status()
        printer("Calibration Status:  ", self.get_cal_status(), "\n")
        results = self.ddr_node.get_property("health_status")
        printer("Overall Health:  ", results["health_status"], "\n")
        if self.is_gen5:
            printer("Crypto Status:  ", self.crypto_type, "\n")
        results = self.ddr_node.get_property("cal_message")
        printer("Message:  ", results["cal_message"], "\n")

        # Check if there is error condition to report
        results = self.ddr_node.get_property("cal_error_msg")
        if results["cal_error_msg"] != "None":
            printer("Error:  ", results["cal_error_msg"], "\n")

        # DDRMC ISR Registers
        printer("\n-------------------\n")
        printer(" Status Registers\n")
        printer("-------------------\n")

        printer("DDRMC ISR Table\n")
        results = self.ddr_node.get_property_group(["ddrmc_isr_e", "ddrmc_isr_w"])
        for key, val in sorted(results.items()):
            printer("  ", key, ":  ", str(val))

        printer("\nUB ISR Table\n")
        results = self.ddr_node.get_property_group(["ub_isr_e", "ub_isr_w"])
        for key, val in sorted(results.items()):
            printer("  ", key, ":  ", str(val))

        # Memory configuration info
        printer("\n----------------------------------\n")
        printer(" Memory Configuration \n")
        printer("----------------------------------\n")
        configs = self.__get_configuration()
        for key, val in configs.items():
            printer(key, ":  ", val)

        # Memory calibration stages info
        printer("\n-----------------------------------\n")
        printer(" Calibration Stages Information \n")
        printer("-----------------------------------\n")
        # Do not do calibration stages if a system error is detected
        sys_error = False
        if self.get_cal_status() == "SYSTEM ERROR":
            sys_error = True
        if not sys_error:
            results = self.get_cal_stages()
            for key, val in sorted(results.items()):
                printer(key, ":  ", val)
        else:
            printer(
                "\nNote: DDRMC system error is detected. Calibration stages decoding will not be provided for ",
                self.name,
                "\n",
            )

        # Cal Margin Analysis
        printer("\n---------------------------------------\n")
        printer(" Calibration Window Margin Analysis \n")
        printer("---------------------------------------\n")

        if not sys_error:
            cal_margin_modes = {}
            left_margins = {}
            right_margins = {}
            center_points = {}
            cal_margin_modes = self.get_cal_margin_mode()
            num_byte = int(configs["Bytes"])
            if not self.is_gen5:
                is_by_8 = True
                num_nibble = int(configs["Nibbles"])
                self.refresh_cal_margin()
                if int(configs["Bits per Byte"]) == 4:
                    is_by_8 = False

                # Main loop to go through dual frequencies
                for freq in range(2):
                    base = "Frequency " + str(freq)
                    # Read Simple
                    key = "f" + str(freq) + "_rd_simp"
                    if cal_margin_modes[key]:
                        # Rising Edge
                        self.__get_cal_margins(
                            key + "_rise", left_margins, right_margins, center_points
                        )
                        printer(
                            "\n",
                            base,
                            " - Read Margin - Simple Pattern - Rising Edge Clock in pS and (delay taps):\n",
                        )
                        self.__output_read_margins(
                            is_by_8,
                            num_byte,
                            num_nibble,
                            left_margins,
                            right_margins,
                            center_points,
                        )
                        # Falling Edge
                        self.__get_cal_margins(
                            key + "_fall", left_margins, right_margins, center_points
                        )
                        printer(
                            "\n",
                            base,
                            " - Read Margin - Simple Pattern - Falling Edge Clock in pS and (delay taps):\n",
                        )
                        self.__output_read_margins(
                            is_by_8,
                            num_byte,
                            num_nibble,
                            left_margins,
                            right_margins,
                            center_points,
                        )
                    # Read Complex
                    key = "f" + str(freq) + "_rd_comp"
                    if cal_margin_modes[key]:
                        # Rising Edge
                        self.__get_cal_margins(
                            key + "_rise", left_margins, right_margins, center_points
                        )
                        printer(
                            "\n",
                            base,
                            " - Read Margin - Complex Pattern - Rising Edge Clock in pS and (delay taps):\n",
                        )
                        self.__output_read_margins(
                            is_by_8,
                            num_byte,
                            num_nibble,
                            left_margins,
                            right_margins,
                            center_points,
                        )
                        # Falling Edge
                        self.__get_cal_margins(
                            key + "_fall", left_margins, right_margins, center_points
                        )
                        printer(
                            "\n",
                            base,
                            " - Read Margin - Complex Pattern - Falling Edge Clock in pS and (delay taps):\n",
                        )
                        self.__output_read_margins(
                            is_by_8,
                            num_byte,
                            num_nibble,
                            left_margins,
                            right_margins,
                            center_points,
                        )
                    # Write Simple
                    key = "f" + str(freq) + "_wr_simp"
                    if cal_margin_modes[key]:
                        self.__get_cal_margins(key, left_margins, right_margins, center_points)
                        printer(
                            "\n",
                            base,
                            " - Write Margin - Simple Pattern - Calibration Window in pS and (delay taps):\n",
                        )
                        self.__output_write_margins(
                            num_byte, left_margins, right_margins, center_points
                        )
                    # Write Complex
                    key = "f" + str(freq) + "_wr_comp"
                    if cal_margin_modes[key]:
                        self.__get_cal_margins(key, left_margins, right_margins, center_points)
                        printer(
                            "\n",
                            base,
                            " - Write Margin - Complex Pattern - Calibration Window in pS and (delay taps):\n",
                        )
                        self.__output_write_margins(
                            num_byte, left_margins, right_margins, center_points
                        )
            # Is Gen5
            else:
                dual_freqs = configs.get("Memory Frequency 1")
                if dual_freqs:
                    num_freqs = 2
                else:
                    num_freqs = 1
                num_ranks = int(configs["Ranks"])
                num_bits_byte = int(configs["Bits per Byte"])
                prbs_ctps_rise = {}
                prbs_ctps_fall = {}
                prbs_ctps = {}

                # Main loop to go through dual frequencies
                for freq in range(num_freqs):
                    base = "Freq " + str(freq) + " Rank "
                    for rank in range(num_ranks):
                        # Read DQ Simp
                        key = "f" + str(freq) + "_rd_dq_simp"
                        if cal_margin_modes[key]:
                            # Rising Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dq_simp_rise"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DQ Margin - Simple Pattern - Rising Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                True,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_rise = center_points.copy()

                            # Falling Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dq_simp_fall"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DQ Margin - Simple Pattern - Falling Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                True,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_fall = center_points.copy()

                        # Read DQ Comp
                        key = "f" + str(freq) + "_rd_dq_comp"
                        if cal_margin_modes[key]:
                            # Rising Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dq_comp_rise"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DQ Margin - Complex Pattern - Rising Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                True,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_rise = center_points.copy()

                            # Falling Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dq_comp_fall"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DQ Margin - Complex Pattern - Falling Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                True,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_fall = center_points.copy()

                        # Read DQ PRBS
                        key = "f" + str(freq) + "_rd_dq_prbs"
                        if cal_margin_modes[key]:
                            # Rising Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dq_prbs_rise"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points, True
                            )
                            if prbs_ctps_rise:
                                center_points = prbs_ctps_rise.copy()
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DQ Margin - PRBS Pattern - Rising Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                True,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_rise.clear()

                            # Falling Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dq_prbs_fall"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points, True
                            )
                            if prbs_ctps_fall:
                                center_points = prbs_ctps_fall.copy()
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DQ Margin - PRBS Pattern - Falling Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                True,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_fall.clear()

                        # Read DBI Simp
                        key = "f" + str(freq) + "_rd_dbi_simp"
                        if cal_margin_modes[key]:
                            # Rising Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dbi_simp_rise"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DBI Margin - Simple Pattern - Rising Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                False,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_rise = center_points.copy()

                            # Falling Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dbi_simp_fall"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DBI Margin - Simple Pattern - Falling Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                False,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_fall = center_points.copy()

                        # Read DBI Comp
                        key = "f" + str(freq) + "_rd_dbi_comp"
                        if cal_margin_modes[key]:
                            # Rising Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dbi_comp_rise"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DBI Margin - Complex Pattern - Rising Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                False,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_rise = center_points.copy()

                            # Falling Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dbi_comp_fall"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DBI Margin - Complex Pattern - Falling Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                False,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_fall = center_points.copy()

                        # Read DBI PRBS
                        key = "f" + str(freq) + "_rd_dbi_prbs"
                        if cal_margin_modes[key]:
                            # Rising Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dbi_prbs_rise"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points, True
                            )
                            if prbs_ctps_rise:
                                center_points = prbs_ctps_rise.copy()
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DBI Margin - PRBS Pattern - Rising Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                False,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_rise.clear()

                            # Falling Edge
                            group_name_base = (
                                "f" + str(freq) + "_rank" + str(rank) + "_rd_dbi_prbs_fall"
                            )
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points, True
                            )
                            if prbs_ctps_fall:
                                center_points = prbs_ctps_fall.copy()
                            printer(
                                "\n",
                                base + str(rank),
                                " - Read DBI Margin - PRBS Pattern - Falling Edge Clock in pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                False,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps_fall.clear()

                        # Write DQ Simp
                        key = "f" + str(freq) + "_wr_dq_simp"
                        if cal_margin_modes[key]:
                            group_name_base = "f" + str(freq) + "_rank" + str(rank) + "_wr_dq_simp"
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Write DQ Margin - Simple Pattern - In pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                True,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps = center_points.copy()

                        # Write DQ Comp
                        key = "f" + str(freq) + "_wr_dq_comp"
                        if cal_margin_modes[key]:
                            group_name_base = "f" + str(freq) + "_rank" + str(rank) + "_wr_dq_comp"
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Write DQ Margin - Complex Pattern - In pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                True,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps = center_points.copy()

                        # Write DQ PRBS
                        key = "f" + str(freq) + "_wr_dq_prbs"
                        if cal_margin_modes[key]:
                            group_name_base = "f" + str(freq) + "_rank" + str(rank) + "_wr_dq_prbs"
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points, True
                            )
                            if prbs_ctps:
                                center_points = prbs_ctps.copy()
                            printer(
                                "\n",
                                base + str(rank),
                                " - Write DQ Margin - PRBS Pattern - In pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                True,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps.clear()

                        # Write DM Simp
                        key = "f" + str(freq) + "_wr_dm_simp"
                        if cal_margin_modes[key]:
                            group_name_base = "f" + str(freq) + "_rank" + str(rank) + "_wr_dm_simp"
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Write DM Margin - Simple Pattern - In pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                False,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps = center_points.copy()

                        # Write DM Comp
                        key = "f" + str(freq) + "_wr_dm_comp"
                        if cal_margin_modes[key]:
                            group_name_base = "f" + str(freq) + "_rank" + str(rank) + "_wr_dm_comp"
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points
                            )
                            printer(
                                "\n",
                                base + str(rank),
                                " - Write DM Margin - Complex Pattern - In pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                False,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
                            prbs_ctps = center_points.copy()

                        # Write DM PRBS
                        key = "f" + str(freq) + "_wr_dm_prbs"
                        if cal_margin_modes[key]:
                            group_name_base = "f" + str(freq) + "_rank" + str(rank) + "_wr_dm_prbs"
                            self.__get_cal_margins(
                                group_name_base, left_margins, right_margins, center_points, True
                            )
                            if prbs_ctps:
                                center_points = prbs_ctps.copy()
                            printer(
                                "\n",
                                base + str(rank),
                                " - Write DM Margin - PRBS Pattern - In pS and (delay taps):\n",
                            )
                            self.__output_gen5_margins(
                                False,
                                num_byte,
                                num_bits_byte,
                                left_margins,
                                right_margins,
                                center_points,
                            )
        else:
            printer(
                "\nNote: DDRMC system error is detected. Margin Analysis will not be provided for ",
                self.name,
                "\n",
            )

        if out_file:
            sys.stdout = orig_out
            out_file.close()

        return

    def set_eye_scan_read_mode(self):
        """
        Set 2D eye scan to read mode

        Args:

        Returns:

        """
        self.ddr_node.set_property({"mgchk_rw_mode": 0})
        self.ddr_node.commit_property_group([])

        return

    def set_eye_scan_write_mode(self):
        """
        Set 2D eye scan to write mode

        Args:

        Returns:

        """
        self.ddr_node.set_property({"mgchk_rw_mode": 1})
        self.ddr_node.commit_property_group([])

        return

    def set_eye_scan_simple_pattern(self):
        """
        Set 2D eye scan traffic pattern to simple mode

        Args:

        Returns:

        """
        self.ddr_node.set_property({"mgchk_pattern_mode": 0})
        self.ddr_node.commit_property_group([])

        return

    def set_eye_scan_complex_pattern(self):
        """
        Set 2D eye scan traffic pattern to complex mode

        Args:

        Returns:

        """
        self.ddr_node.set_property({"mgchk_pattern_mode": 1})
        self.ddr_node.commit_property_group([])

        return

    def select_eye_scan_prbs_pattern(self, choice: int):
        """
        Select 2D eye scan traffic pattern to specific PRBS modes.

        Args:

            choice:
                Choice number valid between 0 and 2 only and corrosponding to following pattern:
                    0) PRBS 23
                    1) PRBS 10
                    2) PRBS 7

        Returns:

        """
        if isinstance(choice, str):
            printer("ERROR: Please enter an integer value from 0 to 2 for PRBS pattern selection.")
            return
        elif choice > 2:
            printer("ERROR: Cannot enter a choice number larger than 2.")
            return

        decoded_pattern_choice = choice + 2
        self.ddr_node.set_property({"mgchk_pattern_mode": decoded_pattern_choice})
        self.ddr_node.commit_property_group([])

        return

    def set_eye_scan_rank(self, rank_num: int):
        """
        Set the target rank number on the DDR for 2D eye scan to run on.
        There will be 4 total possible rank selections. For single slot designs, there will be quad ranks.
        For dual slot designs, there will be dual ranks per slot.
        2D Margin Scan can only be performed on one rank at a time, and not in parallel.

        Args:

            rank_num:
                Rank number valid between 0 and 3 on the DDR for 2D Margin to be run on.

        Returns:

        """
        if isinstance(rank_num, str):
            printer("ERROR: Please enter an integer value from 0 to 3 for Slot/Rank selection.")
            return
        elif rank_num > 3:
            printer("ERROR: Cannot enter a rank_num value larger than 3.")
            return

        self.ddr_node.set_property({"mgchk_rank": rank_num})
        self.ddr_node.commit_property_group([])

        return

    def get_eye_scan_default_read_vref(self) -> int:
        """
        A helper function for Gen5 DDRMC designs for querying default Read VRef value used
        during Calibration, in encoded format for control registers (not in percentage format).

        Args:

        Returns:

            A calculated value of default Read VRef done during Calibration

        """
        results = self.ddr_node.get_property(
            ["mgchk_def_rd_vref_h1p_lsb", "mgchk_def_rd_vref_h1p_msb"]
        )
        vref_lower = results["mgchk_def_rd_vref_h1p_lsb"]
        vref_upper = results["mgchk_def_rd_vref_h1p_msb"]
        vref_value = vref_lower

        if vref_upper > 0:
            vref_value += 512

        return vref_value

    def get_eye_scan_default_write_vref(self) -> int:
        """
        A helper function for Gen5 DDRMC designs for querying default Write VRef value used
        during Calibration, in encoded format for control registers (not in percentage format).

        Args:

        Returns:

            A calculated value of default Write VRef done during Calibration

        """
        results = self.ddr_node.get_property(["mgchk_def_wr_vref"])
        vref_value = results["mgchk_def_wr_vref"]

        return vref_value

    def get_eye_scan_vref_percentage(self, vref: int) -> float:
        """
        A user helper function that takes in an encoding vref value intended for the margin scan,
        and converts it into a percentage value based on the current scan mode
        and memory configuration detected from the current hardware.

        Args:

            vref:
                For Read mode, valid integer range between 0 to 1023
                For Write mode, valid integer range between 0 to 50

        Returns:

            A translated percentage number in floating point

        """
        if isinstance(vref, str):
            printer("ERROR: Please re-enter an integer value for Vref.")
            return
        elif vref > 1023:
            printer("ERROR: Cannot enter VRef values larger than 1023")
            return

        base_value = 0
        incr_size = 0
        ddr4_range_one = 60
        ddr4_range_two = 45
        lp4_range_one = 10
        lp4_range_two = 20.4
        ddr4_incr_size = 0.65
        lp4_incr_size = 0.4
        lp5_base = 10.0
        lp5_incr_size = 0.5
        read_incr_size = 0.0976563
        prop_list = []
        # Gen5 Write Vref
        vref_threshold = 128

        prop_list.append("mgchk_rw_mode")
        prop_list.append("mem_type")
        if not self.is_gen5:
            prop_list.append("mgchk_def_wr_vref_range")
            vref_threshold = 50

        results = self.ddr_node.get_property(prop_list)
        rw_mode = results["mgchk_rw_mode"]
        mem_type = results["mem_type"]
        if not self.is_gen5:
            write_vref_range = results["mgchk_def_wr_vref_range"]

        if rw_mode:
            if vref > vref_threshold:
                printer(
                    "ERROR: Cannot enter Vref values larger than ",
                    vref_threshold,
                    " for Write Margin mode",
                )
                return
            if mem_type == 1:
                if write_vref_range == 0:
                    base_value = ddr4_range_one
                else:
                    base_value = ddr4_range_two
                incr_size = ddr4_incr_size
            elif mem_type == 2:
                if write_vref_range == 0:
                    base_value = lp4_range_one
                else:
                    base_value = lp4_range_two
                incr_size = lp4_incr_size
            elif mem_type == 5:
                base_value = lp5_base
                incr_size = lp5_incr_size
            elif mem_type == 4:
                # This needs to be finalized later
                pass
        else:
            incr_size = read_incr_size

        offset_value = vref * incr_size
        output_vref = base_value + offset_value

        if not rw_mode:
            output_vref = "%.3f" % output_vref
        else:
            output_vref = "%.2f" % output_vref

        return float(output_vref)

    def set_eye_scan_vref_min(self, vref: int) -> float:
        """
        Set 2D eye scan desired minimum vref value to scan

        Args:

            vref:
                For Read mode, valid integer range between 0 to 1023
                For Write mode, valid integer range between 0 to 50

        Returns:

            A translated percentage number in floating point if
            the input VRef is set successfully

        """
        percent_vref = self.get_eye_scan_vref_percentage(vref)

        if percent_vref is not None:
            vref = int(vref)
            self.ddr_node.set_property({"es_vref_min": vref})
            return percent_vref

    def set_eye_scan_vref_max(self, vref: int) -> float:
        """
        Set 2D eye scan desired maximum vref value to scan

        Args:

            vref:
                For Read mode, valid integer range between 0 to 1023
                For Write mode, valid integer range between 0 to 50

        Returns:

            A translated percentage number in floating point if
            the input VRef is set successfully

        """
        percent_vref = self.get_eye_scan_vref_percentage(vref)

        if percent_vref is not None:
            vref = int(vref)
            self.ddr_node.set_property({"es_vref_max": vref})
            return percent_vref

    def set_eye_scan_vref_steps(self, steps: int):
        """
        Set 2D eye scan desired number of vref steps to scan vertically

        Args:

            steps:
                Valid integer range between 1 to 1024 for Read mode.
                Valid integer range between 1 to 51 for Write mode.

        Returns:

        """
        results = self.ddr_node.get_property(["mgchk_rw_mode"])
        rw_mode = results["mgchk_rw_mode"]

        if isinstance(steps, str):
            printer("ERROR: Please re-enter an integer value for Vref steps.")
            return
        elif rw_mode:
            if not self.is_gen5 and (steps > 51):
                printer("ERROR: Cannot enter Vref step sizes larger than 51 for Write Margin mode")
                return
            elif self.is_gen5 and (steps > 129):
                printer("ERROR: Cannot enter Vref step sizes larger than 129 for Write Margin mode")
                return
        elif steps > 1024:
            printer("ERROR: Cannot enter Vref step sizes larger than 1024 for Read Margin mode")
            return

        steps = int(steps)
        self.ddr_node.set_property({"es_vref_steps": steps})

        return

    def __update_eye_scan_data(self):
        edges = ["left", "right"]
        clocks = ["nqtr", "pqtr"]
        clock_defs = {
            "nqtr": "fall_",
            "pqtr": "rise_",
        }
        data_list = []

        engine_data = self.ddr_node.get_eye_scan_data()
        ps_factors = self.ddr_node.get_margin_ps_factors()
        data_length = len(engine_data)

        # If initial vref scan went wrong after enabled, needless to process data further
        if not data_length > 1:
            return

        scan_mode = engine_data[0]["config.mode"]
        nibble_count = engine_data[0]["config.nibbles"]
        byte_count = engine_data[0]["config.bytes"]

        if "Read" in scan_mode:
            old_key_base = "rdmargin_"
            new_key_base = "read_"
            for data in engine_data:
                for nibble in range(int(nibble_count)):
                    for clock in clocks:
                        for edge in edges:
                            old_key = (
                                old_key_base
                                + clock
                                + "_"
                                + edge
                                + "_nibble"
                                + "{:02d}".format(nibble)
                            )
                            ps_name = new_key_base + clock_defs[clock] + "{:02d}".format(nibble)
                            new_key = ps_name + "_" + edge
                            new_key_ps = new_key + "_ps"
                            ps_factor = ps_factors[ps_name]
                            tap_value = data[old_key]
                            ps_value = int(round(tap_value * ps_factor))
                            data[new_key] = tap_value
                            data[new_key_ps] = ps_value
                            del data[old_key]
                temp_data = {key: val for key, val in data.items() if "wrmargin" not in key}
                new_data = OrderedDict(sorted(temp_data.items()))
                data_list.append(new_data)
        else:
            old_key_base = "wrmargin_"
            new_key_base = "write_"
            for data in engine_data:
                for byte in range(int(byte_count)):
                    for edge in edges:
                        old_key = old_key_base + edge + "_byte" + str(byte)
                        ps_name = new_key_base + "{:02d}".format(byte)
                        new_key = ps_name + "_" + edge
                        new_key_ps = new_key + "_ps"
                        ps_factor = ps_factors[ps_name]
                        tap_value = data[old_key]
                        ps_value = int(round(tap_value * ps_factor))
                        data[new_key] = tap_value
                        data[new_key_ps] = ps_value
                        del data[old_key]
                temp_data = {key: val for key, val in data.items() if "rdmargin" not in key}
                new_data = OrderedDict(sorted(temp_data.items()))
                data_list.append(new_data)

        self.eye_scan_data = data_list
        self.ddr_node.set_property({"es_data_need_update": 0})

    def __update_gen5_eye_scan_data(self):
        edges = ["left", "right"]
        clocks = ["nqtr", "pqtr"]
        clock_defs = {
            "nqtr": "fall_",
            "pqtr": "rise_",
        }
        data_list = []

        engine_data = self.ddr_node.get_eye_scan_data()
        data_length = len(engine_data)

        # If initial vref scan went wrong after enabled, needless to process data further
        if not data_length > 1:
            return

        scan_mode = engine_data[0]["config.mode"]
        bit_count = engine_data[0]["config.bits"]
        byte_count = engine_data[0]["config.bytes"]

        if "Read" in scan_mode:
            old_key_base = "rdmargin_"
            new_key_base = "read_"
            for data in engine_data:
                for bit in range(int(bit_count)):
                    for clock in clocks:
                        for edge in edges:
                            old_key = (
                                old_key_base + clock + "_" + edge + "_bit" + "{:02d}".format(bit)
                            )
                            reformat_name = new_key_base + clock_defs[clock] + "{:02d}".format(bit)
                            new_key = reformat_name + "_" + edge
                            tap_value = data[old_key]
                            data[new_key] = tap_value
                            del data[old_key]
                temp_data = {key: val for key, val in data.items() if "wrmargin" not in key}
                new_data = OrderedDict(sorted(temp_data.items()))
                data_list.append(new_data)
        else:
            for data in engine_data:
                temp_data = {key: val for key, val in data.items() if "rdmargin" not in key}
                new_data = OrderedDict(sorted(temp_data.items()))
                data_list.append(new_data)

        self.eye_scan_data = data_list
        self.ddr_node.set_property({"es_data_need_update": 0})

    def __draw_eye_scan_graph(
        self,
        margin_data: List,
        base_name: str,
        margin_title: str,
        def_vrefs: List,
        cur_max: List,
        unit_string: str = "",
        margin_only: bool = False,
    ):
        vrefs = []
        left_margs = []
        right_margs = []
        total_margs = []
        left_ps = []
        right_ps = []
        def_window_found = False
        min_found = False
        max_found = False
        def_margin = 0
        vref_swing = 0.0
        x_min = 0
        x_max = 0
        y_min = 0.0
        y_max = 0.0

        if not self.is_gen5:
            for data in margin_data:
                cur_vref = float(data["config.vrefp"])
                vrefs.append(cur_vref)
                key_name = base_name + "_left"
                left_marg = int(data[key_name])
                left_margs.append(left_marg)
                key_name += "_ps"
                left_ps.append(int(data[key_name]))
                key_name = base_name + "_right"
                right_marg = int(data[key_name])
                right_margs.append(right_marg)
                key_name += "_ps"
                right_ps.append(int(data[key_name]))
                total_margin = left_marg + right_marg
                total_margs.append(total_margin)
                if not def_window_found:
                    if cur_vref == def_vrefs[0]:
                        def_margin = total_margin
                        def_window_found = True
                if cur_vref < def_vrefs[0]:
                    if total_margin == 0:
                        if min_found:
                            min_found = False
                if not min_found:
                    if total_margin > 0:
                        y_min = cur_vref
                        y_max = cur_vref
                        min_found = True
                elif not max_found:
                    if total_margin == 0:
                        max_found = True
                    else:
                        y_max = cur_vref
        else:
            ps_factors = self.ddr_node.get_margin_ps_factors()
            ps_factor = ps_factors[0]

            if "wrmargin" in base_name:
                for data in margin_data:
                    cur_vref = float(data["config.vrefp"])
                    vrefs.append(cur_vref)
                    key_name = base_name + "left_lsb_" + unit_string
                    tap_val = int(data[key_name])
                    key_name = base_name + "left_msb_" + unit_string
                    msb_val = int(data[key_name])
                    left_marg = (msb_val * (2**9)) + tap_val
                    left_margs.append(left_marg)
                    ps_val = int(round(left_marg * ps_factor))
                    left_ps.append(ps_val)
                    key_name = base_name + "right_lsb_" + unit_string
                    tap_val = int(data[key_name])
                    key_name = base_name + "right_msb_" + unit_string
                    msb_val = int(data[key_name])
                    right_marg = (msb_val * (2**9)) + tap_val
                    right_margs.append(right_marg)
                    ps_val = int(round(right_marg * ps_factor))
                    right_ps.append(ps_val)
                    total_margin = left_marg + right_marg
                    total_margs.append(total_margin)
                    if not def_window_found:
                        if cur_vref == def_vrefs[0]:
                            def_margin = total_margin
                            def_window_found = True
                    if cur_vref < def_vrefs[0]:
                        if total_margin == 0:
                            if min_found:
                                min_found = False
                    if not min_found:
                        if total_margin > 0:
                            y_min = cur_vref
                            y_max = cur_vref
                            min_found = True
                    elif not max_found:
                        if total_margin == 0:
                            max_found = True
                        else:
                            y_max = cur_vref
            else:
                for data in margin_data:
                    cur_vref = float(data["config.vrefp"])
                    vrefs.append(cur_vref)
                    key_name = base_name + "_left"
                    left_marg = int(data[key_name])
                    left_margs.append(left_marg)
                    ps_val = int(round(left_marg * ps_factor))
                    left_ps.append(ps_val)
                    key_name = base_name + "_right"
                    right_marg = int(data[key_name])
                    right_margs.append(right_marg)
                    ps_val = int(round(right_marg * ps_factor))
                    right_ps.append(ps_val)
                    total_margin = left_marg + right_marg
                    total_margs.append(total_margin)
                    if not def_window_found:
                        if cur_vref == def_vrefs[0]:
                            def_margin = total_margin
                            def_window_found = True
                    if cur_vref < def_vrefs[0]:
                        if total_margin == 0:
                            if min_found:
                                min_found = False
                    if not min_found:
                        if total_margin > 0:
                            y_min = cur_vref
                            y_max = cur_vref
                            min_found = True
                    elif not max_found:
                        if total_margin == 0:
                            max_found = True
                        else:
                            y_max = cur_vref

        x_min = max(left_ps)
        x_max = max(right_ps)
        max_marg = max(total_margs)
        max_marg_idx = total_margs.index(max_marg)
        max_marg_vref = vrefs[max_marg_idx]
        if not def_window_found:
            def_window = []
            init_list = []
            scan_data = self.eye_scan_data
            init_list.append(scan_data[0])
            self.__draw_eye_scan_graph(
                init_list, base_name, "", def_vrefs, def_window, unit_string, margin_only=True
            )
            def_margin = def_window[0][0]
        vref_swing = round((y_max - y_min), 3)
        cur_max.append(
            (max_marg, max_marg_vref, def_margin, def_vrefs[0], vref_swing, y_min, y_max)
        )

        if margin_only:
            return

        data_defs = {
            "VRef": vrefs,
            "Left_Margin": left_margs,
            "Right_Margin": right_margs,
            "Left_PS": left_ps,
            "Right_PS": right_ps,
        }

        df = pd.DataFrame(data_defs)
        fig = go.Figure()

        for col in df.columns[1:2]:
            df_left = df[col]
            ps_left = df[df.columns[3]]
            data_left = list(zip(df_left, ps_left))
            fig.add_trace(
                go.Bar(
                    x=-ps_left.values,
                    y=df[df.columns[0]],
                    orientation="h",
                    name=col,
                    customdata=data_left,
                    marker_color="#75ad0a",
                    hovertemplate="Taps: %{customdata[0]} <br>pS: %{customdata[1]} <br>VRef: %{y}",
                )
            )

        for col in df.columns[2:3]:
            df_right = df[col]
            ps_right = df[df.columns[4]]
            data_right = list(zip(df_right, ps_right))
            fig.add_trace(
                go.Bar(
                    x=ps_right,
                    y=df[df.columns[0]],
                    orientation="h",
                    name=col,
                    customdata=data_right,
                    marker_color="#0073e6",
                    hovertemplate="Taps: %{customdata[0]} <br>pS: %{customdata[1]} <br>VRef: %{y}",
                )
            )

        if len(def_vrefs) > 1:
            display_vref = def_vrefs[1]
            display_text = "Cal. VRef: " + str(display_vref)
        else:
            display_vref = def_vrefs[0]
            display_text = "Def. VRef: " + str(display_vref)

        fig.add_trace(
            go.Scatter(
                x=[-x_min, x_max],
                y=[display_vref, display_vref],
                mode="lines",
                line=dict(
                    color="#92b1a9",
                    width=4,
                    dash="dot",
                ),
                name=display_text,
                customdata=[x_min, x_max],
                hovertemplate="%{customdata}",
            )
        )

        fig.update_layout(
            barmode="relative",
            height=600,
            width=900,
            bargap=0.01,
            plot_bgcolor="#0f263e",
            title=margin_title,
            xaxis=dict(title="Delay in pS"),
            yaxis=dict(title="VRef Percentage"),
            legend_orientation="v",
            legend_x=1.0,
            legend_y=0.0,
        )

        return fig

    def save_eye_scan_data(self, file_name=None):
        """
        Save the eye scan 2D margin data from the most recent run to a file in CSV format

        Args:

            file_name: the file name or full path for the data file to be saved.
                       A default file name will be given if none is supplied.

        Returns:

        """
        out_file = None

        if not file_name:
            # Gives a default for users
            file_name = "eye_scan_data"

        file_name = os.path.splitext(file_name)[0] + ".csv"
        if self.is_gen5:
            self.__update_gen5_eye_scan_data()
        else:
            self.__update_eye_scan_data()

        if not self.eye_scan_data:
            printer("ERROR: No scan data is found. Please re-run 2D eye scan.")
            return

        data_list = self.eye_scan_data
        all_keys = list(data_list[0].keys())

        with open(file_name, "w", newline="") as out_file:
            data_writer = csv.DictWriter(out_file, all_keys)
            data_writer.writeheader()
            data_writer.writerows(data_list)
            printer("\nInfo: Data are being formatted and saved to:", out_file.name)

    def load_eye_scan_data(self, file_name):
        """
        Load the eye scan 2D margin data from a data file

        Args:

            file_name: the file name or full path for the data file to be read and load

        Returns:

        """
        file_path = Path(file_name)
        data_list = []

        with file_path.open(mode="r") as data_file:
            reader = csv.DictReader(data_file)
            data_reader = list(reader)
            for data in data_reader:
                data_list.append(data)
            file_name = data_file.name

        self.eye_scan_data = data_list
        self.ddr_node.set_property({"es_data_need_update": 0})
        printer("\nInfo: Scan data are being loaded successfully from:", file_name)

    def run_eye_scan(self, done: request.DoneFutureCallback = None) -> bool or request.CsFuture:
        """
        Kick off a run on 2D margin analysis and prepare scan data which is needed to draw an
        eye scan graph. The default scan settings will be used if users have not configured with
        new settings. See the following related commands for more info:

            save_eye_scan_data()
            load_eye_scan_data()
            display_eye_scan()
            check_eye_scan_window()
            check_eye_scan_height()

        Args:
            done: Optional command callback that will be invoked when the scan is finished

        Returns:
            Boolean value indicating success of the 2D margin scan

        """

        # Perform some logical data check on user settings
        prop_list = []
        prop_list.append("mgchk_rw_mode")
        prop_list.append("es_vref_min")
        prop_list.append("es_vref_max")
        prop_list.append("es_vref_steps")

        results = self.ddr_node.get_property(prop_list)
        rw_mode = results["mgchk_rw_mode"]
        vref_min = results["es_vref_min"]
        vref_max = results["es_vref_max"]
        vref_steps = results["es_vref_steps"]

        if rw_mode:
            if self.is_gen5:
                vref_threshold = 128
            else:
                vref_threshold = 50

            if (vref_min > vref_threshold) or (vref_max > vref_threshold):
                printer(
                    "ERROR: Cannot set Vref values larger than ",
                    vref_threshold,
                    " under Write margin mode.",
                )
                return False
            if vref_steps > vref_threshold + 1:
                printer(
                    "ERROR: Cannot set Vref step sizes larger than ",
                    vref_threshold + 1,
                    " for Write Margin mode",
                )
                return False

        if vref_min > vref_max:
            printer("ERROR: Cannot set minimum Vref value larger than maximum Vref value.")
            return False

        percent_min = self.get_eye_scan_vref_percentage(vref_min)
        percent_max = self.get_eye_scan_vref_percentage(vref_max)
        printer("Min VRef is set at:", percent_min, "%, encoded value", vref_min)
        printer("Max VRef is set at:", percent_max, "%, encoded value", vref_max)
        printer("Number of VRef steps is set at:", vref_steps)

        vref_incr_size = int((vref_max - vref_min) / (vref_steps - 1))
        max_step_size = vref_max - vref_min + 1

        if vref_incr_size < 1:
            printer(
                "ERROR: For the selected VRef Min and VRef Max, the maximum step size is: ",
                str(max_step_size),
            )
            printer(
                "Please either reduce the VRef steps or increase the VRef range by adjusting Min/Max."
            )
            return False

        scan_future = request.CsFutureSync(done)

        # Institute a Progress Bar
        progress_bar = PercentProgressBar()
        in_progress = "[bold dodger_blue2]In Progress[/]"
        progress_bar.add_task(
            description="2D Margin Scan Progress",
            status=PercentProgressBar.Status.STARTING,
            visible=True,
        )

        def progress_update(future):
            progress_bar.update(completed=future.progress * 100, status=in_progress)

        def done_scan(future):
            if future.error is not None:
                progress_bar.update(status=PercentProgressBar.Status.ABORTED)
            else:
                progress_bar.update(completed=100, status=PercentProgressBar.Status.DONE)

            if future.error:
                scan_future.set_exception(future.error)
            else:
                scan_future.set_result(None)

        def finalize(future):
            # Call on main thread for any final processing desired
            result_list = future._result
            if result_list[0]:
                # Can think about if want to provide auto-draw or
                # auto-save log for the users
                status = "COMPLETED"
            else:
                status = "ERROR"

            printer("2D margin scan run status is: ", status)

        self.ddr_node.future(
            progress=progress_update, done=done_scan, final=finalize
        ).run_eye_scan_async(data_checked=True)
        printer("Info: Margin Scans in progress...")

        return scan_future if done else scan_future.result

    def display_eye_scan(
        self,
        unit_index: int = 0,
        return_as_list: bool = False,
        display_type: str = "dynamic",
        get_margin_only: bool = False,
    ):
        """
        Assemble and display 2D eye scan drawing in chart format. By default, it tries to process and
        draw from the scan data found in the most recent 2D eye scan run. Users have the option to
        load scan data first prior to using this function in order to display specific set of scan
        data, see load_eye_scan_data().

        By default, if no unit_index is specified, unit zero from the scan data will be displayed.
        For read margins, both rising and falling edge clock scan data will be processed and displayed.

        Args:
            unit_index: Specify the index from a set of scan data users intend to display, based on
                unit mode found in the scan settings (bit, nibble, byte)

            return_as_list: Optional argument, default to False. If set to True, the function will
                not display graphs by default. Instead, it returns a list of the grapghing objects
                back to the caller.

            display_type: Optional argument, default to dynamic. If set to static, a static image
                will be returned. If set to dynamic, a dynamic, interactive javascript view will be
                returned.

            get_margin_only: Optional argument, default to False. If set to True, this API call
                will process scan data only, per the specified unit_index, find its widest margin
                windows and return the values and along with their VRef points as a list of tuples.
                No graphing results will be given.

        Returns:
            A list of Figure objects from Plotly, if return_as_list is specified as True,
            None is return otherwise by default, but graphs will be displayed in a browser.

            If optional argument get_margin_only is set to True. A list, which contains tuples as
            (int, float) representing the widest window margin found and along with their VRef points
            after a scan is returned.

            For Read Margin, two data results are displayed or returned. (First data for Rising
            Clock edge, and the second for Falling Clock edge) For Write Margin, only one data
            result will be displayed or returned.

        """
        if (not get_margin_only) and (not _plotting_pkgs_available):
            raise ImportError(
                f"Plotting packages not installed! Please run ",
                "'pip install chipscopy[plotly,pandas]'",
            )

        result = self.ddr_node.get_property(["es_data_need_update"])
        data_need_update = result["es_data_need_update"]

        if data_need_update:
            self.eye_scan_data.clear()
            if self.is_gen5:
                self.__update_gen5_eye_scan_data()
            else:
                self.__update_eye_scan_data()

        if not self.eye_scan_data:
            printer(
                "ERROR: No scan data is found. Please re-run 2D eye scan, "
                "or load scan data first from a data log file."
            )
            return

        if isinstance(unit_index, str):
            printer("ERROR: Please re-enter an integer value for unit_index.")
            return

        df_vrefs = []
        figure_list = []
        data_list = self.eye_scan_data
        mc_id = data_list[0]["config.mc_id"]
        scan_mode = data_list[0]["config.mode"]
        unit_mode = data_list[0]["config.unit"]
        rank_num = data_list[0]["config.rank"]
        mem_type = data_list[0]["config.mem_type"]
        df_vref = data_list[0]["config.df_vrefp"]
        df_vref = float(df_vref)
        df_vrefs.append(df_vref)
        if not self.is_gen5:
            nibble_count = data_list[0]["config.nibbles"]
        else:
            bit_count = data_list[0]["config.bits"]
        byte_count = data_list[0]["config.bytes"]
        unit_val = "{:02d}".format(unit_index)
        unit_str = ""
        max_window = []  # list of tuple (int, float)
        window_results = []  # list of tuple (int, float)

        # Some logical data check first on user inputs
        if unit_mode == "bit":
            if (unit_index < 0) or (unit_index >= int(bit_count)):
                max_unit = int(bit_count) - 1
                err_msg = (
                    "ERROR: unit_index given for "
                    + unit_mode
                    + " mode needs to be between 0 and "
                    + str(max_unit)
                )
                err_msg += ", based on current scan data found."
                printer(err_msg)
                return
            unit_str = unit_mode + unit_val
            if "Read" in scan_mode:
                key_name = "read_vref_val_" + unit_str
                cal_vref = data_list[0].get(key_name)
            else:
                key_name = "write_vref_val_" + unit_str
                cal_vref = data_list[0].get(key_name)
            if cal_vref:
                cal_vref = float(cal_vref)
                df_vrefs.append(cal_vref)
        elif unit_mode == "nibble":
            if (unit_index < 0) or (unit_index >= int(nibble_count)):
                max_unit = int(nibble_count) - 1
                err_msg = (
                    "ERROR: unit_index given for "
                    + unit_mode
                    + " mode needs to be between 0 and "
                    + str(max_unit)
                )
                err_msg += ", based on current scan data found."
                printer(err_msg)
                return
        elif unit_mode == "byte":
            if (unit_index < 0) or (unit_index >= int(byte_count)):
                max_unit = int(byte_count) - 1
                err_msg = (
                    "ERROR: unit_index given for "
                    + unit_mode
                    + " mode needs to be between 0 and "
                    + str(max_unit)
                )
                err_msg += ", based on current scan data found."
                printer(err_msg)
                return
            unit_str = unit_mode + str(unit_index)

        mc_name = mc_id
        data_list = data_list[1:]

        if "Read" in scan_mode:
            clocks = ["rise_", "fall_"]
            clock_defs = {
                "rise_": " - Rising Edge Clock - ",
                "fall_": " - Falling Edge Clock - ",
            }

            for clock in clocks:
                name_base = "read_" + clock + unit_val
                margin_mode = (
                    scan_mode
                    + clock_defs[clock]
                    + unit_mode.capitalize()
                    + str(unit_index)
                    + "<br>"
                )
                margin_mode += mc_name + " - "
                margin_mode += mem_type + " - Rank" + str(rank_num)
                if not get_margin_only:
                    fig = self.__draw_eye_scan_graph(
                        data_list, name_base, margin_mode, df_vrefs, max_window
                    )
                    figure_list.append(fig)
                    printer(
                        "Margin window found:",
                        clock_defs[clock],
                        max_window[0][0],
                        "fine taps at VRef %:",
                        max_window[0][1],
                    )
                else:
                    self.__draw_eye_scan_graph(
                        data_list, name_base, margin_mode, df_vrefs, max_window, margin_only=True
                    )
                    window_results.append(max_window[0])
                max_window.clear()

        else:
            margin_mode = scan_mode + " - " + unit_mode.capitalize() + str(unit_index) + "<br>"
            margin_mode += mc_name + " - "
            margin_mode += mem_type + " - Rank" + str(rank_num)
            if not self.is_gen5:
                name_base = "write_" + unit_val
                if not get_margin_only:
                    fig = self.__draw_eye_scan_graph(
                        data_list, name_base, margin_mode, df_vrefs, max_window
                    )
                    printer(
                        "Margin window found:",
                        max_window[0][0],
                        "fine taps at VRef %:",
                        max_window[0][1],
                    )
                    figure_list.append(fig)
                else:
                    self.__draw_eye_scan_graph(
                        data_list, name_base, margin_mode, df_vrefs, max_window, margin_only=True
                    )
                    window_results.append(max_window[0])
            else:
                name_base = "wrmargin_"
                if not get_margin_only:
                    fig = self.__draw_eye_scan_graph(
                        data_list, name_base, margin_mode, df_vrefs, max_window, unit_str
                    )
                    printer(
                        "Margin window found:",
                        max_window[0][0],
                        "fine taps at VRef %:",
                        max_window[0][1],
                    )
                    figure_list.append(fig)
                else:
                    self.__draw_eye_scan_graph(
                        data_list,
                        name_base,
                        margin_mode,
                        df_vrefs,
                        max_window,
                        unit_str,
                        margin_only=True,
                    )
                    window_results.append(max_window[0])

        if get_margin_only:
            return window_results

        if not return_as_list:
            for fig in figure_list:
                if display_type == "dynamic":
                    fig.show()
                elif display_type == "static":
                    if not _jupyter_available:
                        raise ImportError(
                            f"Jupyter packages not installed! Please run 'pip install chipscopy[jupyter]'"
                        )
                    image_bytes = fig.to_image(format="png")
                    ipython_image = Image(image_bytes)
                    display(ipython_image)
                else:
                    raise ValueError("display_type argument must be static or dynamic")
        else:
            return figure_list

    def check_eye_scan_height(self, unit_index: int = 0):
        """
        Calculate and obtain the VRef swing/height found from a margin scan by a specified unit index.
        The height is defined by having the lowest point of VRef that a non-zero margin window starts, and ends
        with the highest VRef point found in the same fashion. By default, it tries to process from the scan
        data found in the most recent 2D eye scan run. Users have the option to load scan data first prior to
        using this function, in order to process specific set of scan data, see load_eye_scan_data().

        Args:
            unit_index: Specify the index from a set of scan data users intend to display, based on
                unit mode found in the scan settings (bit, nibble, byte)

        Returns:
            A list, which contains lists of floating point values representing the VRef percentage values
            in the following item order: [VRef swing, minimum VRef, maximum VRef]
            For Read Margin, two sets of data list are being returned. (First item from the list is for Rising
            Clock edge, and the second is for Falling Clock edge) For Write Margin, only one set of data list
            will be returned.
        """
        vref_results = []  # List of list [float, float, float]

        window_results = self.display_eye_scan(unit_index, get_margin_only=True)

        for window in window_results:
            vref_results.append([window[4], window[5], window[6]])

        return vref_results

    def check_eye_scan_window(self, unit_index: int = 0, default_vref: bool = False):
        """
        Calculate and obtain the widest 2D eye scan margin window by a specified unit index. This function
        also checks the margin window found at default VRef setting determined by the memory controller.
        By default, it tries to process from the scan data found in the most recent 2D eye scan run. Users
        have the option to load scan data first prior to using this function, in order to process specific
        set of scan data, see load_eye_scan_data().

        Args:
            unit_index: Specify the index from a set of scan data users intend to display, based on
                unit mode found in the scan settings (bit, nibble, byte)

            default_vref: Optional argument, default to False, which means the function will check and report
                the widest margin window found and along with its corrosponding vref value at the earliest point.
                If set to True, the function will report margin window found at its default VRef point instead.

        Returns:
            A list, which contains tuples of (int, float) will be returned. The first integer item from the
            tuples represents the window margin found after a scan, and the second item indicates the VRef
            setting value in floating point.

            For Read Margin, two data results are being returned. (First tuple of data is for Rising
            Clock edge, and the second is for Falling Clock edge) For Write Margin, only one tuple
            result will be returned.

        """
        checked_results = []  # List of tuples (int, float)
        return_idx0 = 0
        return_idx1 = 1

        window_results = self.display_eye_scan(unit_index, get_margin_only=True)

        if default_vref:
            return_idx0 += 2
            return_idx1 += 2

        # Read Modes
        if len(window_results) > 1:
            for window in window_results:
                checked_results.append((window[return_idx0], window[return_idx1]))
        else:
            # Write Modes
            window = window_results[0]
            checked_results.append((window[return_idx0], window[return_idx1]))

        return checked_results

    def get_cal_vref(
        self,
        unit_index: int = 0,
    ):
        """
        Only supported on Gen5 DDRs. This function finds and returns the calibrated VRef in percentage
        value per specified unit index. By default, it tries to process from the scan data found in the
        most recent 2D eye scan run. Users have the option to load scan data first prior to calling this
        function, in order to process specific set of scan data, see load_eye_scan_data().

        Args:
            unit_index: Specify the target index within a set of scan data, based on unit mode found in
            the scan settings (bit, nibble, byte)

        Returns:
            A floating point number that represents the calibrated VRef in percentage value.
            Note that this number will be defaulted to zero if the calibrated VRef cannot be found.

        """

        if not self.is_gen5:
            printer("ERROR: This function is currently supported on Gen5 DDRs only.")
            return

        result = self.ddr_node.get_property(["es_data_need_update"])
        data_need_update = result["es_data_need_update"]

        if data_need_update:
            self.eye_scan_data.clear()
            self.__update_gen5_eye_scan_data()

        if not self.eye_scan_data:
            printer(
                "ERROR: No scan data is found. Please re-run 2D eye scan, "
                "or load scan data first from a data log file."
            )
            return

        if isinstance(unit_index, str):
            printer("ERROR: Please re-enter an integer value for unit_index.")
            return

        data_list = self.eye_scan_data
        scan_mode = data_list[0]["config.mode"]
        unit_mode = data_list[0]["config.unit"]
        bit_count = data_list[0]["config.bits"]
        unit_val = "{:02d}".format(unit_index)
        unit_str = ""

        # Logical data check first on user input
        if unit_mode == "bit":
            if (unit_index < 0) or (unit_index >= int(bit_count)):
                max_unit = int(bit_count) - 1
                err_msg = (
                    "ERROR: unit_index given for "
                    + unit_mode
                    + " mode needs to be between 0 and "
                    + str(max_unit)
                )
                err_msg += ", based on current scan data found."
                printer(err_msg)
                return
            unit_str = unit_mode + unit_val
            if "Read" in scan_mode:
                key_name = "read_vref_val_" + unit_str
                cal_vref = data_list[0].get(key_name)
            else:
                key_name = "write_vref_val_" + unit_str
                cal_vref = data_list[0].get(key_name)
        else:
            printer(
                "ERROR: Scan data shows that non-bit unit mode found, "
                "which is currently not supported."
            )
            return

        if cal_vref:
            cal_vref = float(cal_vref)
        else:
            cal_vref = 0.0

        return cal_vref

    def __str__(self):
        return self.name



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Set, Dict
from dataclasses import dataclass

from chipscopy.api import CoreType
from chipscopy.api._detail.debug_core import DebugCore

# TODO: PerfMonTrafficQOSClass should have an alias on the top level
from chipscopy.dm.harden.noc_perfmon.traffic_classes import PerfMonTrafficQOSClass
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.client.noc_perfmon_core_client import NoCPerfMonCoreClient
from chipscopy.utils import printer


@dataclass
class NocPerfmon(DebugCore["NoCPerfMonCoreClient"]):
    """NocPerfMon is the top level API for accessing NocPerfmon.
    It wraps the old client noc_perfmon_core_client.NoCPerfMonCoreClient(CoreClient) class
    and simplifies the use model by hiding tcf and node lower level details.
    """

    def __init__(self, noc_tcf_node):
        super(NocPerfmon, self).__init__(CoreType.NOC_ROOT, noc_tcf_node)

        self.name = noc_tcf_node.props["Name"]

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name}

    def __str__(self):
        return self.name

    def initialize(self, done: DoneHWCommand = None):  # pragma: no cover
        """
        This method collects the clocking information for client use later in the configuration of this service. It
        also refreshes the client's view of specific properties for the root node *ONLY*.

        This call must be invoked before clocking information is requested from the device

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            No data is returned directly from this method.

        """
        self.core_tcf_node.initialize(done)

    def discover_noc_elements(self, done: DoneHWCommand = None) -> List[str]:
        """

        WARNING - This function is deprecated and will be removed in the future

        This method dynamically scans the NoC to determine the active nodes in the programmed design. If no
        nodes are programmed or the NPI domain is inaccessible this will return an empty list.

        This list will serve as the set from which elements may be chosen for runtime performance analysis.

        Unrouted, disabled NoC elements cannot be configured for measurement. It may be assumed their counters would
        always produce zero count values.

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            Returns a list of active NMU, NSU/DDRMC_NOC, and NPS elements. Unrouted, disabled elements are not returned
            and cannot be monitored.

        """
        element_list = self.core_tcf_node.discover_noc_elements(done)
        return element_list

    def configure_monitors(
        self,
        monitors: Set[str],
        sampling_intervals: dict,
        traffic_class: PerfMonTrafficQOSClass,
        sample_count: int,
        done: DoneHWCommand = None,
        extended_monitor_config: dict = None,
    ):
        """
        This method configures and enables the performance monitoring on the parameter set of elements. Upon completion
        of the configuration, data will start to be emitted asynchronously from ChipScope Server back to the requesting
        client. The client must implement an Event Listener to properly receive the data.

        Each element will be configured to monitor both read and write traffic independently. QOS classes of a
        particular direction of traffic will be applied to the monitor responsible for that type of traffic
        ``read|write``.

        Args:
            monitors (List[str]):
                List of NoC elements to enable to stream performance data.
                For Example -
                ``['NOC_NMU128_X0Y0', 'DDRMC_X0Y0', ...]``
            sampling_intervals (dict):
                The requested ``period_ms`` for sampling. The client must ensure this is a supported period.
                **NOTE**: this structure needs to contain information for NPI, NoC and each DDRMC controller clk
                NPI, NoC are sampling periods (ms) (floats) and DDRMC controller clk domain is a tap (int)
                -- for reference see the XSA class in the utilities to understand how the ctrl clk works.
            traffic_class (PerfMonTrafficQOSClass):
                The traffic classes bitwise-or'd that shall be monitored.
            sample_count (int):
                *OPTIONAL* The number of intervals or windows to capture. 1 - will configure supporting hardware for
                single shot. 2-int_max will configure all hardware for continuous mode and will stop streaming data when
                the requested number of samples have been sent to the client. -1 is the default and means the streaming
                will run indefinitely.
            filter_spec (dict[int]):
                Will move! hold off on using this for now.
                *OPTIONAL* Dictionary defining the AXI filtering to use. If this parameter is not supplied then no
                filtering will occur and all transactions will be counted by the samplers. No validation is done on
                these fields, they will be masked and shifted, as they are received. For convenience any fields not
                specified will retain the default value of zero, except max fields (AxSIZE, AxQoS, AxLEN) which have the
                largest supported values). The supported fields include

                ``AxPROT[3]``, ``AxBURST[2]``, ``AxSIZE_MIN[3]``, ``AxLEN_MIN[8]``, ``AxID[16]``
                ``TDEST[10]`` (AXIS only), ``DST_ID[12]``, ``AxLOCK[1]``, ``AxQoS_MIN[4]``, ``AxCACHE[4]``,
                ``AxSIZE_MAX[3]``, ``AxQoS_MAX[4]``, ``AxLEN_MAX[8]``

            done:
                Optional command callback that will be invoked when the response is received.

            extended_monitor_config:
                Optional argument that is a dictionary of additional monitor configuration. The keys must correspond to
                the monitors. "tslide" is the only supported key at this time. Tslide will change the meaning of the
                latency counters for the specified monitor (read and write channel). The meaning of the counter values
                can be represented by 2^(tslide) clocks. Default tslide is 0, corresponding to 1 clock per count. In
                practice this is a tradeoff between precision and range. When the counters overflow error flags will be
                set in the associated metrics. If overflow is observed, the tslide should be increased to avoid that
                condition. Otherwise there will be a "clipping" in the data. Note, only NMU and NSU's support the tslide
                extended configuraiton option. DDRMC does not, and will ignore the value, if supplied by the client.

        Returns:
            No data is returned from this call.

        """
        self.core_tcf_node.configure_monitors(
            monitors=monitors,
            sampling_intervals=sampling_intervals,
            traffic_class=traffic_class,
            sample_count=sample_count,
            done=done,
            extended_monitor_config=extended_monitor_config,
        )

    def get_supported_sampling_periods(
        self,
        ref_clk_feq_mhz=33.3333,
        ddrmc_freq_mhz={"ddrmc_main_0": 800.0},
        done: DoneHWCommand = None,
    ) -> Dict[str, List[float]]:
        """
        Get the supported sampling periods (ms) for various clock domains based on input clocks and design parameters.
        This function queries the PLLs of the device to determine actual clock speeds of locked PLLs.

        Args:
            done: Optional command callback that will be invoked when the response is received.
            ddrmc_freq_mhz (dict[str,float]): dictionary of DDRMC instance names as the keys and the value is the MC clock in MHz float.
            ref_clk_feq_mhz (float): ref_clk input frequency in MHz, value can be attained from .csa or .hwh files. Default is 33.333MHz

        Returns:
            Dictionary with two top level keys ``NoC``, ``NPI``, and memory subsystem components whose values are lists
            of supported sampling periods measured in milliseconds (float).

        """
        sampling_dict = self.core_tcf_node.get_supported_sampling_periods(
            ref_clk_feq_mhz, ddrmc_freq_mhz, done
        )
        return sampling_dict

    def get_clk_info(self, done: DoneHWCommand = None) -> dict:  # pragma: no cover
        """
        DEPRECATED will remove in 2022.1

        Reports NPI and NoC clock information from the running design. These are parameters set by the designer in
        Vivado.

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            A dictionary with two top level keys ``NoC`` and ``NPI`` whose values are dictionaries that
            detail the ``freq_mhz``, frequency in MHz (float) and the ``period_ms``, period in milliseconds (float).

        """
        frequency_dict = self.core_tcf_node.get_clk_info(done)
        return frequency_dict

    def enumerate_noc_elements(
        self, scan_nodes, done: DoneHWCommand = None, raw_mode=False
    ) -> List[str]:
        results = self.core_tcf_node.enumerate_noc_elements(scan_nodes, done)
        disabled = results["disabled"]
        invalid = results["invalid"]
        if len(disabled) != 0:
            printer(
                "Warning: These nodes were requested but are disabled in the hardware, verify the design:"
            )
            for d in disabled:
                printer(f"  {d}")
        if len(invalid) != 0:
            printer(
                "Warning: These nodes were requested but are not valid for "
                "this hardware, double check the device and the site names in "
                "the design:"
            )
            for i in invalid:
                printer(f"  {i}")
        if raw_mode:
            return results
        else:
            return results["enabled"]



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Callable, List, NamedTuple, Optional, Tuple
from chipscopy.client.axis_ila_core_client import ILATriggerCondition, ILAPort

#
#  IlaMapper -  Mapping match compare values to port match units.
#
#    Look at test_axis_ila_mapper.py, to see data structure json dumps of input data
#    and output data.
#


@dataclass
class Port:
    index: int
    mu_count: int
    width: int
    # 'mu_values' filled out by mapper
    mu_values: [str] = field(default_factory=lambda: [])


@dataclass
class Match:
    name: str
    port_index: int
    # Position of match value LSB, on the port.
    bit_index: int
    op: str
    value: str
    # Un-mapped value is -1.
    mu_index: int = -1
    # 'source' fill out by mapper
    source: str = ""
    # Corresponds to all probes names added to this Match. For debugging.
    _source_names: List[str] = field(default_factory=lambda: [])
    # _source_names: [str] = None
    # Compare value indices in same order as _source_names. For debugging.
    _source_value_indices: List[int] = field(default_factory=lambda: [])
    # _source_value_indices: [int] = None

    @staticmethod
    def get_sort_key():
        return lambda match: (match.port_index << 16) + match.bit_index

    def is_dont_care(self):
        return all(ch == "X" for ch in self.value)

    def ensure_proper_types(self):
        """
        Some old Vivado_lab C++ JSON examples have str values instead of ints, convert to int.
        Ensure values are upper case.
        """
        self.value = self.value.upper()
        if type(self.bit_index) == str:
            self.bit_index = int(self.bit_index)
        if type(self.port_index) == str:
            self.port_index = int(self.port_index)
        if not self._source_names:
            self._source_names = []
        if not self._source_value_indices:
            self._source_value_indices = []

    def update_source(self, force: bool = True) -> None:
        if self.source and not force:
            return
        self.source = " ".join(
            f"{name}<{idx}>" for name, idx in zip(self._source_names, self._source_value_indices)
        )
        # To be compatible with old C++ generated json
        self.source += " "

    def flip_one_bit_neq_to_eq(self) -> None:
        if self.op != "!=" or len(self.value) != 1:
            return
        # Have a "1=" one bit value, which can be flipped.
        self.op = "=="
        self.value = negate_bit_value(self.value)
        # Mark name, for debug printout
        self.name = "~" + self.name

    def merge_shared_ports(self, other_match) -> object:
        """
        if feasible, will merge with 'other' Match, putting merged result into this Match.
        Caller need to remove obsolete 'other'.
        Return merged Match object. None if no merge.
        """
        m1: Match = other_match
        if self.port_index != m1.port_index or self.op != "==" or m1.op != "==":
            return None

        merged_value = self.make_merged_value(self, m1)
        if not merged_value:
            return None

        merged = Match(
            self.name + "&" + m1.name,
            self.port_index,
            min(self.bit_index, m1.bit_index),
            self.op,
            merged_value,
        )

        merged._source_names = self._source_names
        merged._source_names.extend(m1._source_names)
        merged._source_value_indices = self._source_value_indices
        merged._source_value_indices.extend(m1._source_value_indices)
        return merged

    @staticmethod
    def make_merged_value(match0, match1) -> str:
        # Returns empty string, if merge was not possible.

        def merge_ch(ch0: str, ch1: str) -> str:
            if ch0 == ch1 or ch1 == "X":
                return ch0
            elif ch0 == "X":
                return ch1
            else:
                return ""

        m0: Match = match0
        m1: Match = match1
        new_bit_index = min(m0.bit_index, m1.bit_index)
        new_len = max(m0.bit_index + len(m0.value), m1.bit_index + len(m1.value)) - new_bit_index
        # Normalize values to start at the new bit_index and have same length.
        m0_value = Match.resize_value(m0.value, m0.bit_index, new_bit_index, new_len)
        m1_value = Match.resize_value(m1.value, m1.bit_index, new_bit_index, new_len)
        merged_value = [merge_ch(ch0, ch1) for ch0, ch1 in zip(m0_value, m1_value)]
        if all(merged_value):
            return "".join(merged_value)
        else:
            return ""

    @staticmethod
    def resize_value(
        mu_value: str, mu_value_bit_index: int, new_bit_index: int, new_width: int
    ) -> str:
        """
        A value is grown by filling out with don't-care 'X' in the beginning and/or the end.
        Note! In a string value the LSB is at the end of the string.

        Args:
            mu_value (str): Old value to re-sized.
            mu_value_bit_index (int): LSB bit location on port.
            new_bit_index (int): LSB bit of new value.
            new_width (int): Bit width of the new value.

        Returns: Upsized value.

        """
        if (new_bit_index > mu_value_bit_index) or (
            new_bit_index + new_width < mu_value_bit_index + len(mu_value)
        ):
            raise ValueError("Internal Error: Match unit value of the wrong length.")

        value = mu_value + "X" * (mu_value_bit_index - new_bit_index)
        value = "X" * (new_width - len(value)) + value
        return value

    def make_port_match_value(self, port_width: int) -> str:
        value = Match.resize_value(self.value, self.bit_index, 0, port_width)
        if len(value) < port_width:
            raise ValueError("Internal Error: Wrong match value length.")
        return self.op + value


class Condition:
    def __init__(self, op: str, name: str = "", matches: [Match] = None):
        self.name = name
        self.op: str = op
        self.matches: [Match] = matches if matches else []

    def __str__(self):
        return f"\nop: {self.op}\nmatches: {self.matches}"

    def merge_shared_ports(self) -> None:
        # Only allowed to merge match values when 'op' is 'and' or 'nand'.
        if self.op != "and" and self.op != "nand":
            return

        res = []
        work_list = self.matches
        while len(work_list) > 1:
            remaining = []
            match0, work_list = work_list[0], work_list[1:]
            for match2 in work_list:
                # Try merging  match value with the rest of the work list.
                merged = match0.merge_shared_ports(match2)
                if merged:
                    match0 = merged
                else:
                    # Keep match2.
                    remaining.append(match2)

            res.append(match0)
            work_list = remaining

        # work_list may still have an element.
        res.extend(work_list)
        self.matches = res


class IlaMapperEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Match):
            obj.update_source()

        if isinstance(obj, (Port, Match, Condition, IlaMapper)):
            obj_dict = obj.__dict__
            obj_dict.pop("__module__", None)
            obj_dict.pop("__class__", None)
            if isinstance(obj, Match):
                # Get rid of internal data members
                obj_dict.pop("_source_names", None)
                obj_dict.pop("_source_value_indices", None)

            return obj_dict
        return json.JSONEncoder.default(self, obj)


class IlaMapper:
    def __init__(self, ports=None, conditions=None, is_basic_capture=False):
        self.ports: [Port] = ports if ports else []
        self.conditions: [Condition] = conditions if conditions else []
        self._is_basic_capture = is_basic_capture

    def __str__(self):
        return to_json(self)

    def add_ports(self, port_width_with_match_unit_count: [(int, int)]) -> None:
        self.ports = [
            Port(index=index, width=p_width_mu_count[0], mu_count=p_width_mu_count[1])
            for index, p_width_mu_count in enumerate(port_width_with_match_unit_count)
        ]

    def add_condition(self, op: str, name: str = "") -> int:
        """
        Add new condition. Return index of condition.

        Args:
            op (): Lower case operator string, e.g.'and', 'or', 'nor', 'nand'
            name (): Empty string, or ILA core register.

        Returns: Condition index, for added item.

        """
        self.conditions.append(Condition(op, name))
        return len(self.conditions) - 1

    def add_probe_value(
        self,
        condition_index: int,
        probe_name: str,
        port_index: int,
        bit_index: int,
        op: str,
        value: str,
    ) -> None:
        upper_no_underscore_value = "".join(ch if ch != "_" else "" for ch in value.upper())
        match = Match(probe_name, port_index, bit_index, op, upper_no_underscore_value)
        cond: Condition = self.conditions[condition_index]
        cond.matches.append(match)

    def merge_shared_ports(self) -> None:
        for cond in self.conditions:
            cond.merge_shared_ports()

    def flip_one_bit_neq_to_eq(self) -> None:
        for cond in self.conditions:
            for match in cond.matches:
                match.flip_one_bit_neq_to_eq()

    def remove_dont_care_matches(self) -> None:
        for cond in self.conditions:
            cond.matches = [match for match in cond.matches if not match.is_dont_care()]

    def sort_matches(self):
        for cond in self.conditions:
            cond.matches.sort(key=Match.get_sort_key())

    def ensure_proper_types(self) -> None:
        for cond in self.conditions:
            for match in cond.matches:
                match.ensure_proper_types()

    def assign_match_source_info(self) -> None:
        tally = defaultdict(int)
        for cond in self.conditions:
            for match in cond.matches:
                count = tally[match.name]
                match._source_names.append(match.name)
                match._source_value_indices.append(count)

    def map_match_units(self) -> bool:
        """
        Returns True if trigger/capture values fit into port match units.
        """
        result = [self.map_match_unit(mu) for cond in self.conditions for mu in cond.matches]
        return all(result)

    def map_match_unit(self, match: Match) -> bool:
        if match.port_index > len(self.ports):
            raise ValueError(
                f"Internal Error: Match Unit {match.name} refers to non"
                f"-existent port index {match.port_index}"
            )
        port = self.ports[match.port_index]
        value = match.make_port_match_value(port.width)
        if value in port.mu_values:
            # Re-use actual port match unit.
            match.mu_index = port.mu_values.index(value)
        else:
            match.mu_index = len(port.mu_values)
            port.mu_values.append(value)
        # Report if we ran out of port match units.
        return match.mu_index < port.mu_count

    def report_overflow(self) -> [str]:
        def bin_to_hex_if_possible(value: str) -> str:
            if all(ch in ("0", "1") for ch in value):
                return hex(int(value, 2))
            else:
                return value

        def get_compare_values_str(matches: [Match]) -> str:
            compare_values = [
                f"\t{mm.name} {mm.op} {bin_to_hex_if_possible(mm.value)}" for mm in matches
            ]
            return "\n".join(compare_values)

        # Create {port: [match]} map
        port_to_map = defaultdict(list)
        for cond in self.conditions:
            for match in cond.matches:
                port_to_map[match.port_index].append(match)

        overflows = []
        for p_index, port in enumerate(self.ports):
            if port.mu_count < len(port_to_map[p_index]):
                msg = get_compare_values_str(port_to_map[p_index])
                overflows.append(msg)
        return overflows

    __CONDITION_MAPPING_OP = {
        ILATriggerCondition.AND.name.lower(): "==",
        ILATriggerCondition.NAND.name.lower(): "!=",
        ILATriggerCondition.OR.name.lower(): "||",
        # 'nor' -> 'and' after deMorgan
        ILATriggerCondition.NOR.name.lower(): "==",
    }

    @staticmethod
    def get_unused_tc_value(match_unit_count: int) -> str:
        return "==" + ("x" * match_unit_count)

    def create_test_get_mu_index_function(self) -> (int, Callable[[int, int], int]):
        """
        Create function, based on self.ports.
        Intended to be used only in tests.
        Function assumes that match units are paired with Ports in port index order.

        Returns: Tuple( match_unit_count,
                        Function which can be used in test examples when calling "get_mapping()")
        """
        pp_pmu_list = [
            (pp.index, pmu_index) for pp in self.ports for pmu_index in range(pp.mu_count)
        ]
        mu_index_by_port_index_and_mu_index_with_port = {
            pp_pmu: mu_index for mu_index, pp_pmu in enumerate(pp_pmu_list)
        }

        def get_mu_index_fn(port_index: int, mu_index_within_port: int) -> int:
            return mu_index_by_port_index_and_mu_index_with_port.get(
                (port_index, mu_index_within_port), -1
            )

        return len(pp_pmu_list), get_mu_index_fn

    def get_reg_mu_mapping(
        self, mu_count: int, get_mu_index: Callable[[int, int], int]
    ) -> {str: str}:
        """

        Args:
            mu_count: Total ILA match unit count,
            get_mu_index(Callable): A function given (port_index, match_index_within_port)
                returns the ILA match unit index

        Returns: Dict {"reg_name":match value} for Trigger conditions and match units.
        """
        res = {}

        # TCs (trigger conditions)
        for c_index, condition in enumerate(self.conditions):
            mu_used = [False] * mu_count
            for match in condition.matches:
                if match.mu_index >= 0:
                    mu_used[get_mu_index(match.port_index, match.mu_index)] = True
            # LSB should be at end for value.
            mu_used.reverse()
            if condition.op == "nor":
                # Nor handled by deMorgan. Invert inputs using '0'. nor -> and
                cmp_value = ["0" if used else "X" for used in mu_used]
            else:
                cmp_value = ["1" if used else "X" for used in mu_used]
            if condition.name:
                condition_name = condition.name
            else:
                condition_name = "sqc" if self._is_basic_capture else "tc" + str(c_index)
            res[condition_name] = IlaMapper.__CONDITION_MAPPING_OP[condition.op] + "".join(
                cmp_value
            )

        # Match Units
        for p_index, port in enumerate(self.ports):
            for mu_index_for_port, mu_val in enumerate(port.mu_values):
                mu_index = get_mu_index(p_index, mu_index_for_port)
                res["mu" + str(mu_index)] = mu_val

        return res

    def map(self, skip_overflow_messages=True) -> (bool, Optional[str]):
        """
        Map match values onto port match units.

        Args:
            skip_overflow_messages (bool): If True,

        Returns: A tuple - all_mapped_status, None or List of overflow messages.

        """
        self.ensure_proper_types()
        self.assign_match_source_info()
        self.sort_matches()
        self.remove_dont_care_matches()
        self.flip_one_bit_neq_to_eq()
        self.merge_shared_ports()
        all_mapped = self.map_match_units()
        overflow_msg = None
        if not all_mapped and not skip_overflow_messages:
            overflow_msg = self.report_overflow()
        return all_mapped, overflow_msg


def to_json(ila_mapper: IlaMapper, skip_unused_ports=True) -> str:
    if skip_unused_ports:
        ports = [port for port in ila_mapper.ports if port.mu_values]
        ila_mapper1 = IlaMapper(ports, ila_mapper.conditions)
    else:
        ila_mapper1 = ila_mapper
    res = json.dumps(ila_mapper1, cls=IlaMapperEncoder, indent=2)
    return res


def from_json(j_str: str) -> IlaMapper:
    def _decode_condition(cond: dict) -> Condition:
        matches = cond.get("matches", None)
        if matches and type(matches) == list:
            cond["matches"] = [Match(**mm) for mm in matches]
        return Condition(**cond)

    data_tree = json.loads(j_str)
    # Re-creating list of dataclass Port.
    ports = data_tree.get("ports", None)
    if ports and type(ports) == list:
        ports = [Port(index, **pp) for index, pp in enumerate(ports)]
    conditions = data_tree.get("conditions", None)
    # Re-creating 'condition' list.
    if conditions and type(conditions) == list:
        conditions = [_decode_condition(cond) for cond in conditions]
    res = IlaMapper(ports, conditions)
    return res


# Todo: Change MapPortSeq to PortSegment? or PortRange
class MapPortSeq(NamedTuple):
    """A port bit-range."""

    port_idx: int
    high_idx: int
    low_idx: int

    def __str__(self):
        if self.high_idx == self.low_idx:
            return f"port{self.port_idx}[{self.high_idx}]"
        else:
            return f"port{self.port_idx}[{self.high_idx}:{self.low_idx}]"

    @staticmethod
    def to_str(seqs: []) -> str:
        return " ".join(s.__str__() for s in seqs)


_map_re = re.compile(r"(\d+)(\[(\d+)(:(\d+))?\])?")


def port_map_str_to_port_seqs(
    name: str, map_str: str, ports: Optional[List[ILAPort]] = None
) -> [MapPortSeq]:
    """
        If ports is None, some checks are turned off.

    port<int>
    port<int>[<int>]
    port<int>[<int>:<int>]
    """

    def _to_int(s: str):
        return int(s) if s else None

    def _parse_map_item(item: str) -> Tuple[int, int, int]:
        """
        If ports is None, some checks are turned off.
        Return port_idx, high_bit_idx, low_bit_idx.
        """
        if item.startswith("port"):
            s = item[len("port") :]
        elif item.startswith("probe"):
            s = item[len("probe") :]
        else:
            raise ValueError(f'Probe "{name}" map item "{item}" does not start with prefix "port".')

        match = _map_re.match(s)
        if not match:
            raise ValueError(f'Probe "{name}" map item "{item}" has incorrect syntax.')

        port_str, _, left_str, __, right_str = match.groups()
        port_idx, left, right = int(port_str), _to_int(left_str), _to_int(right_str)
        if ports:
            if (port_idx < 0) or (port_idx >= len(ports)):
                raise ValueError(
                    f'Probe "{name}" map item "{item}" has invalid port index "{port_idx}"'
                    f" valid values are 0 to {len(ports)}."
                )
        if right is None:
            # port<int>[<int>] syntax used, if no "right index"
            high, low = left, left
        else:
            high, low = left, right

        port_width = None
        if ports:
            port_width = ports[port_idx]["bit_width"]
            if low is None and high is None:
                # port < int > syntax used. No range specified
                if port_width == 1:
                    high, low = 0, 0
                else:
                    raise ValueError(
                        f'Probe "{name}" map item "{item}" needs explict range '
                        f'"[<high>:<low>]" since port bit width is larger than one.'
                    )

        if low < 0:
            raise ValueError(
                f'Probe "{name}" map item "{item}" has bit range index less than zero.'
            )

        if high < low:
            raise ValueError(
                f'Probe "{name}" map item "{item}" high range index "{high}" '
                f'is less than low bit index "{low}".'
            )

        if port_width is not None and (port_width <= high):
            raise ValueError(
                f'Probe "{name}" map item "{item}" high range index "{high}" '
                f'is invalid. Port has "{port_width}" bit(s).'
            )
        return MapPortSeq(port_idx, high, low)

    items = map_str.lower().split()
    map_seqs = [_parse_map_item(item) for item in items]
    return map_seqs


def negate_bit_value(bit_ch: str) -> str:
    negate_dict = {
        "0": "1",
        "1": "0",
        "X": "X",
        "R": "L",
        "F": "S",
        "T": "N",
        "B": "N",
        "N": "T",
        "L": "R",
        "S": "F",
    }
    return negate_dict.get(bit_ch.upper(), "X")



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass, field
from typing import Optional, List

from chipscopy.shared.ila_data import TraceInfo


EXTERNAL_TRACE_DATA_ALIGNMENT_BIT_COUNT = 512
"""Raw data samples are aligned on 64 byte boundaries."""


@dataclass
class DataSegment:
    """
    DataSegment refer to data section of raw data consisting of consecutive data records,
    or refers to one gap record in the raw data.
    """

    start: int
    end: int
    sample_count: int
    is_gap: bool


@dataclass
class DataWindow:
    data: memoryview
    last_record_offset: int
    """ Points to next address after record."""
    sample_count: int
    last_gap_count: int
    segments: List[DataSegment] = field(default_factory=lambda: [])


def decode_external_data(
    data: bytearray, trace_info: TraceInfo, beyond_last_record_addrs: [int], last_gap_counts: [int]
) -> (bytearray, bool):
    # Input data records are aligned on 64 byte boundary.
    # MSB, in record, is record type. 1 for gap record. 0 got regular sample record.
    # Returns: (data formatted for ILAWaveform.data, "data has gap flag")

    abc = EXTERNAL_TRACE_DATA_ALIGNMENT_BIT_COUNT
    align_byte_count = abc // 8
    in_sample_byte_size = ((trace_info.data_width + 1 + abc - 1) // abc) * align_byte_count
    out_sample_byte_size = (trace_info.data_width + 1 + 7) // 8
    gap_block_size = 1024 if trace_info.window_size >= 1024 else trace_info.window_size
    gap_sample = None

    def get_gap_sample_count(window_data: memoryview, offset: int) -> int:
        # offset points to the just beyond the record.
        # sample "block" count is in the lower 32-bits
        read_count = int.from_bytes(
            window_data[offset - in_sample_byte_size : offset - in_sample_byte_size + 4],
            byteorder="little",
            signed=False,
        )

        return read_count * 1024

    def unroll_window(dw: DataWindow) -> None:
        # The resulting unroll-ed window has the last record at the end of the window buffer.
        # If any gaps, the beginning of the buffer is garbage, until real records begin.
        if dw.last_record_offset == len(data):
            return
        buf = bytearray(len(dw.data))
        buf[0 : len(dw.data) - dw.last_record_offset] = data[dw.last_record_offset :]
        buf[len(dw.data) - dw.last_record_offset :] = dw.data[: dw.last_record_offset]
        data[:] = buf[:]

    def find_data_segments(dw: DataWindow) -> None:
        samples_to_process = dw.sample_count

        def flush_data_segment(start_offset: int, end_offset: Optional[int]) -> None:
            if start_offset is None or end_offset is None:
                return None
            seg_sample_count = (end_offset - start_offset) // in_sample_byte_size
            seg = DataSegment(start_offset, end_offset, seg_sample_count, is_gap=False)
            dw.segments.append(seg)
            return None

        if dw.last_gap_count:
            # Insert any gaps after last written record to DDR.
            seg = DataSegment(None, None, dw.last_gap_count, is_gap=True)
            dw.segments.append(seg)
            samples_to_process -= dw.last_gap_count

        current_data_end = None
        current_record_start = None
        for offset in range(len(dw.data), 0, -in_sample_byte_size):
            if samples_to_process <= 0:
                break
            current_record_start = offset - in_sample_byte_size
            # Gap records has msb set to '1'.
            is_gap = dw.data[offset - 1] & 0x80 != 0
            if is_gap:
                current_data_end = flush_data_segment(offset, current_data_end)
                gap_sample_count = get_gap_sample_count(dw.data, offset)
                seg = DataSegment(current_record_start, offset, gap_sample_count, is_gap)
                dw.segments.append(seg)
                samples_to_process -= gap_sample_count
            else:
                if current_data_end is None:
                    current_data_end = offset
                samples_to_process -= 1

        flush_data_segment(current_record_start, current_data_end)
        if dw.segments is not None:
            dw.segments.reverse()

        # Handle too many gap samples at in the beginning of the window.
        if samples_to_process < 0 and dw.segments and dw.segments[0].is_gap:
            if dw.segments[0].sample_count + samples_to_process > 0:
                dw.segments[0].sample_count += samples_to_process

        if samples_to_process > 0:
            raise ValueError("Waveform data is corrupted. Did not find all samples in window.")

    def order_samples_expand_gaps(dw: DataWindow) -> None:
        # Put data segments first in the window.
        nonlocal gap_sample
        to_offset = 0
        for seg in dw.segments:
            if seg.is_gap:
                if gap_sample is None:
                    gap_sample = (1 << trace_info.data_width).to_bytes(
                        in_sample_byte_size, byteorder="little", signed=False
                    )
                next_to_offset = to_offset + seg.sample_count * in_sample_byte_size
                dw.data[to_offset:next_to_offset] = gap_sample * seg.sample_count
            else:
                next_to_offset = to_offset + seg.end - seg.start
                dw.data[to_offset:next_to_offset] = dw.data[seg.start : seg.end]
            to_offset = next_to_offset

    def remove_sample_high_bytes(data: bytearray, total_sample_count: int) -> bytearray:
        if out_sample_byte_size == in_sample_byte_size:
            return data

        # Remove most significant bytes, from each sample.
        move_len = out_sample_byte_size
        to_addr = 0
        for sample_index, from_addr in enumerate(range(0, len(data), in_sample_byte_size)):
            if sample_index >= total_sample_count:
                break
            data[to_addr : to_addr + move_len] = data[from_addr : from_addr + move_len]
            to_addr += move_len
        return data[:to_addr]

    #
    data_windows: List[DataWindow] = []
    total_sample_count = 0
    window_byte_count = in_sample_byte_size * trace_info.window_size
    data_mv = memoryview(data)
    for window_index, window_addr in enumerate(range(0, len(data), window_byte_count)):
        if trace_info.partial_window_sample_count and window_index == trace_info.window_count - 1:
            samples_in_window = trace_info.partial_window_sample_count
        else:
            samples_in_window = trace_info.window_size
        total_sample_count += samples_in_window

        window_memory = data_mv[window_addr : window_addr + window_byte_count]
        beyond_last_record_addr = beyond_last_record_addrs[window_index] - window_addr
        last_gap_count = last_gap_counts[window_index]
        if beyond_last_record_addr == 0:
            beyond_last_record_addr = window_byte_count
        data_windows.append(
            DataWindow(window_memory, beyond_last_record_addr, samples_in_window, last_gap_count)
        )

    for data_window in data_windows:
        unroll_window(data_window)
        find_data_segments(data_window)
        order_samples_expand_gaps(data_window)

    res = remove_sample_high_bytes(data, total_sample_count)
    return res, gap_sample is not None



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from enum import Enum, unique
from pprint import pformat
from typing import Optional, Callable, List, Set

from chipscopy import dm
from chipscopy.api._detail import dataclass_fields, filter_props
from chipscopy.dm import NodeListener


@unique
class CoreType(Enum):
    IBERT = "IBERT"
    NOC_ROOT = "NOC CORE"
    DDR = "DDRMC"
    HBM = "HBM"
    AXIS_ILA = "AXIS ILA"
    AXIS_VIO = "AXIS VIO"
    AXIS_PCIE = "AXIS PCIE"
    AXIS_TRACE = "AXIS TRACE"
    AXI_DEBUG_HUB = "AXI_DEBUG HUB"
    SYSMON = "SYSMON"


@dataclass(frozen=True)
class CoreInfo:
    """Debug Fabric Core version information."""

    core_type: int
    """Core type number."""
    drv_ver: int
    """SW/IP interface version."""
    core_major_ver: int
    """Core IP major version."""
    core_minor_ver: int
    """Core IP minor version."""
    tool_major_ver: int
    """Major version of the design tool, which created the core."""
    tool_minor_ver: int
    """Minor version of the design tool, which created the core."""
    uuid: str
    """Core UUID. UUID is unique among debug cores in the design."""

    def __str__(self) -> str:
        return pformat(self.__dict__, 2)


CORE_INFO_MEMBERS = dataclass_fields(CoreInfo)


def get_core_info(tcf_node: dm.Node) -> Optional[CoreInfo]:
    if set(CORE_INFO_MEMBERS).issubset(tcf_node.props):
        # Only fabric cores have core info information.
        # Convert UUID to a hex string.
        ci = filter_props(tcf_node.props, CORE_INFO_MEMBERS)
        if isinstance(ci["uuid"], bytearray):
            uuid = ci["uuid"].copy()
            uuid.reverse()
            ci["uuid"] = uuid.hex().upper()
        # pytcf is returning this as a bytes object rather than byte array
        elif isinstance(ci["uuid"], bytes):
            uuid = ci["uuid"]
            ci["uuid"] = "".join(reversed([f"{x:02X}" for x in uuid]))
        return CoreInfo(**ci)

    return None


class DMNodeListener(NodeListener):
    """
    The Session hooks into the hw_server and cs_server with a DMNodeListener to
    listen to events that happen when hardware changes occur.
    This enables the session to keep track of asynchronous events to the device
    and adjust cached values as needed to remain consistent with the hardware
    state.

    When an events happen on a tracked node, the registered callback is called
    with node event details.

    Usage:

    ::

        def _node_callback(self, action: DMNodeListener.NodeAction, node: Node, props: Optional[Set]):
            ... handle callback ...


        # register for callbacks
        self.hw_server.get_view("memory").add_node_listener(DMNodeListener(self._node_callback))

        ...


    """

    @unique
    class NodeAction(Enum):
        NODE_ADDED = "NODE_ADDED"
        NODE_REMOVED = "NODE_REMOVED"
        NODE_CHANGED = "NODE_CHANGED"

    def __init__(self, callback: Callable[[NodeAction, dm.Node, Optional[Set]], None]):
        super().__init__()
        self._callback = callback

    def nodes_added(self, nodes: List[dm.Node]):
        if self._callback:
            for node in nodes:
                self._callback(DMNodeListener.NodeAction.NODE_ADDED, node, None)

    def node_changed(self, node: dm.Node, updated_keys: Set[str]):
        if self._callback:
            self._callback(DMNodeListener.NodeAction.NODE_CHANGED, node, updated_keys)

    def nodes_removed(self, nodes: List[dm.Node]):
        if self._callback:
            for node in nodes:
                self._callback(DMNodeListener.NodeAction.NODE_REMOVED, node, None)
